/*
 * Licensed Materials - Property of IBM Corp.
 * IBM UrbanCode Build
 * IBM UrbanCode Deploy
 * IBM UrbanCode Release
 * IBM AnthillPro
 * (c) Copyright IBM Corporation 2002, 2014. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by
 * GSA ADP Schedule Contract with IBM Corp.
 */
package com.urbancode.anthill3.plugin.subversion;

import java.text.DecimalFormatSymbols;
import java.util.Formatter;
import java.util.Locale;

public class PerfLogger {

    //**************************************************************************
    // CLASS
    //**************************************************************************

    static public enum Level {
        ALL, TRACE, DEBUG, INFO, WARN, ERROR, FATAL, OFF
    }

    static public Level DEFAULT_LEVEL = Level.OFF;
    static {
        String levelEnv = System.getenv()['SVN_PLUGIN_LOG_LEVEL'];
        if (levelEnv) {
            try {
                DEFAULT_LEVEL = Level.valueOf(levelEnv.toUpperCase())
            }
            catch (IllegalArgumentException e) {
                // unknown level, ignore
            }
        }
    }

    //**************************************************************************
    // INSTANCE
    //**************************************************************************

    final PrintStream log;
    final String category;
    Level level = DEFAULT_LEVEL;
    long startTime = System.nanoTime();

    public PerfLogger(PrintStream log, String category) {
        this.log = log;
        this.category = category;
    }

    public void reset() {
        startTime = System.nanoTime();
    }

    public void trace(String message) {
        log(Level.TRACE, message);
    }

    public void trace(String message, Throwable t) {
        log(Level.TRACE, message, t);
    }

    public void info(String message) {
        log(Level.INFO, message);
    }

    public void info(String message, Throwable t) {
        log(Level.INFO, message, t);
    }

    public void debug(String message) {
        log(Level.DEBUG, message);
    }

    public void debug(String message, Throwable t) {
        log(Level.DEBUG, message, t);
    }

    public void warn(String message) {
        log(Level.WARN, message);
    }

    public void warn(String message, Throwable t) {
        log(Level.WARN, message, t);
    }

    public void log(Level lvl, String message) {
        if (isEnabledFor(lvl)) {
            log.println(format(message));
        }
    }

    public void log(Level lvl, String message, Throwable t) {
        if (isEnabledFor(lvl)) {
            log.println(format(message)+": "+t);
            e.printStackTrace(log);
            log.flush();
        }
    }

    public boolean isDebugEnabled() {
        return isEnabledFor(Level.DEBUG);
    }

    public boolean isEnabledFor(Level lvl) {
        return level <= lvl;
    }

    public String format(String message) {
        long duration = System.nanoTime() - startTime;
        return format0(message, duration);
    }

    /**
     * Format the log message content
     * @param message
     * @param nanos
     * @return
     */
    private String format0(String message, long nanos) {
        Formatter formatter = new Formatter(Locale.US);
        try {
            long millis = nanos / 1000000L;
            char decimal = '.';
            Locale locale = formatter.locale();
            if (locale != null) {
                DecimalFormatSymbols symbols = new DecimalFormatSymbols(locale);
                decimal = symbols.getDecimalSeparator();
            }
            formatter.format("[%s, duration=%,d%c%03d] %s",
                    category,
                    (long)(millis / 1000L),
                    decimal,
                    (int)(millis % 1000L),
                    message);
            return formatter.toString();
        }
        finally {
            formatter.close();
        }
    }
}
