package com.urbancode.anthill3.plugin.subversion

import com.urbancode.anthill3.*
import com.urbancode.anthill3.plugin.subversion.changelog.*;

import java.util.Date;
import java.util.TimeZone;
import java.util.regex.Pattern;
import java.text.SimpleDateFormat

import groovy.util.slurpersupport.GPathResult;
import groovy.util.XmlSlurper;


public class SCMChangelog extends SCMStep {

    //**************************************************************************
    // CLASS
    //**************************************************************************

    //**************************************************************************
    // INSTANCE
    //**************************************************************************

    // changelog boundaries
    Date startDate;
    Date endDate;
    String startRevision;
    String endRevision;

    // filters for changeset
    final ChangeSetFilter changeSetFilter = new ChangeSetFilter();

    /**
     *
     * @return xml document of changes
     */
    public def execute() {

        if (this.revision) {
            // this should be a redundant with a check in the step script
            String msg = "Skipping changelog of revision based checkout for %s:r%s";
            println(String.format(msg, targetUrl, revision));
            return;
        }

        //
        // Get Change Log
        //

        def logOutput = runLogCommand()

        //
        // Parse Change Log into changesets
        //

        // the result change set data
        List<ChangeSet> changeSets = parseLogIntoChangeSets(logOutput);

        //
        // Construct AHP XML message and upload to AHP
        //

        if (changeSets) {
            ChangeSetXMLHelper xmlHelper = new ChangeSetXMLHelper()
            xmlHelper.repoType = REPO_TYPE;
            return xmlHelper.toXML(changeSets);
        }
        else {
            println "No changes detected"
            return null
        }
    }

    /**
     * Parse an input of changelog command into a list of AHP ChangeSets
     * @param logOutput output from a {@link #runLogCommand()} method
     * @return List of ahp changesets
     */
    public List<ChangeSet> parseLogIntoChangeSets(String logOutput) {
        List<ChangeSet> changeSets = [];

        if (!logOutput) {
            // no results
            return changeSets;
        }
        initSvnInfo(); // need svn info to properly parse log paths

        String moduleName = null;
        def modulePattern = Pattern.compile('([^/]+)/?$')
        def moduleMatcher = modulePattern.matcher(repoRoot)
        if (moduleMatcher.find()) {
            moduleName = moduleMatcher.group(1);
        }

        new XmlSlurper().parseText(logOutput).'logentry'.each{ GPathResult svnchangeset ->
            def svnChangesetDateText = svnchangeset.'date'[0].text()
            svnChangesetDateText = svnChangesetDateText.replaceFirst('[0-9]{3}Z$', 'Z')
            def svnDate = getSvnDateOutFormat().parse(svnChangesetDateText) // 2009-05-06T20:12:41.284343Z

            ChangeSet changeSet = new ChangeSet();
            changeSet.module = moduleName;
            changeSet.id = svnchangeset.@revision.text();
            changeSet.user = svnchangeset.'author'[0].text();
            changeSet.date = svnDate;
            changeSet.message = svnchangeset.'msg'[0].text();

            svnchangeset.'paths'[0].'path'.each{ path ->
                ChangeSetFileEntry fileEntry = new ChangeSetFileEntry();
                fileEntry.revision = changeSet.id;
                fileEntry.type = path.@action.text();
                fileEntry.path = path.text();
                changeSet.fileSet << fileEntry;
            }
            changeSets << changeSet;
        }

        // filter results
        changeSets.retainAll{ ChangeSet changeSet ->
            if (startDate && changeSet.date < startDate || endDate && endDate < changeSet.date) {
                // outside of date range
                return false
            }

            boolean allowAuthor = changeSet.hasAllowedAuthor(changeSetFilter);
            boolean allowAnyPaths = changeSet.hasAllowedPath(changeSetFilter);
            if (allowAuthor && allowAnyPaths) {
                // retain changeset
                return true;
            }
            else {
                def message = new StringBuilder("Changeset ${changeSet.id} skipped because ")
                if (!allowAuthor) {
                    message << "it has excluded author ${changeSet.user}"
                }
                if (!allowAuthor && !allowAnyPaths) {
                    message << " and "
                }
                if (!allowAnyPaths) {
                    message << "it contains only excluded file paths (${changeSet.fileSet.collect{it.path}})"
                }
                println message
                // exclude changeset
                return false;
            }
        }

        return changeSets;
    }

    protected String runLogCommand() {
        // QP should use Dates, CL uses revisions
        if (!startRevision && !startDate) {
            println "No start point specified; skipping changelog"
            return ""
        }

        def logCommand = getBaseCmd();
        logCommand << 'log'
        logCommand << '--verbose'
        logCommand << '--xml'
        String revArg = startRevision ?: ('{'+formatChangelogDate(startDate)+'}')
        if (endRevision){
            revArg += ':'+endRevision
        }
        else if (endDate) {
            revArg += ':'+'{'+formatChangelogDate(endDate)+'}'
        }
        logCommand << '--revision' << revArg
        logCommand << targetUrl;

        println "Checking for changes"
        def logOutput = getCommandText(logCommand);
        return logOutput;
    }

    protected String formatChangelogDate(Date date) {
        def df = this.getDateArgumentFormatter()
        return date ? df.format(date) : null;
    }
}
