package com.urbancode.anthill3.plugin.subversion;

import com.urbancode.anthill3.*

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.DateFormat;
import java.util.Date;
import java.util.regex.Matcher;

abstract public class SCMStep {

    //**************************************************************************
    // CLASS
    //**************************************************************************

    static public final String REPO_TYPE = 'svn'

    // DateFormat for passing dates to svn as arguments
    static public DateFormat getSvnDateFormat() {
        final def GMT = TimeZone.getTimeZone("GMT")
        final def SVN_DATE  = new java.text.SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S Z")
        SVN_DATE.timeZone = GMT
        return SVN_DATE;
    }

    // format svn uses in its output
    static public DateFormat getSvnDateOutFormat() {
        final def GMT = TimeZone.getTimeZone("GMT")
        final def SVN_DATE_OUT_FORMAT  = new java.text.SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.S'Z'")
        SVN_DATE_OUT_FORMAT.timeZone = GMT
        return SVN_DATE_OUT_FORMAT;
    }

    static public DateFormat getAhpDateFormat() {
        final def GMT = TimeZone.getTimeZone("GMT")
        final def AHP_DATE_FORMAT = new java.text.SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S Z");
        AHP_DATE_FORMAT.timeZone = GMT
        return AHP_DATE_FORMAT;
    }

    /*
     * Use of null (e.g. \0 OR %x00) breaks git log format string under msysgit,
     * so we will use the unit-separator character as an alternative to null when needed.
     */
    static protected final char UNIT_SEPARATOR = '\037'
    static protected final String UNIT_SEPARATOR_FMT = '%x1F'

    //**************************************************************************
    // INSTANCE
    //**************************************************************************

    def out = System.out;

    String scmCommand;
    File directory;

    String svnBaseUrl;
    String svnUser;
    String svnPass;
    boolean noAuthCache = false
    boolean trustServerCert = false;

    String revision; // use a specific revision
    Date date;       // use a specific date?
    String checkoutPath; // the path to checkout the WC from
    String tagsPath      // tags directory
    String branchesPath; // branches directory
    @Deprecated
    String fromTag; // use a particular tag
    @Deprecated
    String fromBranch; // use a particular branch

    // SVN INFO
    String currentRevision
    String lastChangedRevision
    String wcUrl;
    String repoRoot;

    /**
     * Initialize the workspace if needed
     */
    protected void initWorkspace() {

        // create the working directory
        directory.mkdirs()
        if (!directory.exists()) {
            throw new Exception("Could not create working directory ${workDir}")
        }
        else if (directory.isFile()) {
            throw new Exception("Specified working directory is a file!! ${workDir}")
        }
    }

    protected void initSvnInfo() {
        if (repoRoot) {
            // don't eval if we already have it
            return;
        }
        String targetUrl = this.targetUrl; // avoid computing multiple times in this method

        // run svn info on the workspace and set the revision as a property on the build life
        def infoCommand = getBaseCmd();
        infoCommand << 'info'
        if (revision) {
            infoCommand << '--revision' << revision
        }
        else if (date) {
            infoCommand << '--revision' << ('{'+getSvnDateFormat().format(date)+'}')
        }
        else {
            // Default svn behavior is the PEG revision of the url. (which in turn defaults to HEAD for remote operations)
            // We are deviating from this because we did not know better.
            infoCommand << '--revision' << 'HEAD'
        }
        infoCommand << targetUrl;

        println("Getting source info")
        def infoOutput = getCommandText(infoCommand);
        currentRevision = infoOutput.find(~/(?mi)^\s*Revision:\s*(.*)$/) {match, rev -> return rev }
        lastChangedRevision = infoOutput.find(~/(?mi)^\s*Last Changed Rev:\s*(.*)$/) {match, rev -> return rev}
        repoRoot = infoOutput.find(~/(?mi)^\s*Repository Root:\s*(.*)$/) {match, root -> return root }

        // Using value of svn-info URL
        //  probably we should add PEG of revision to ensure validity of path discovered.
        //wcUrl = infoOutput.find(~/(?mi)^\s*URL:\s*(.*)$/) {match, url -> return url + "@${currentRevision}"}
        wcUrl = targetUrl;

        assert currentRevision : 'Could not determine Revision';
        assert lastChangedRevision : 'Could not determine Last Changed Revision';
        assert wcUrl : 'Could not determine URL';
        assert repoRoot : 'Could not determine Repository Root';
    }

    public String getTargetUrl() {
        if (wcUrl != null) {
            return wcUrl;
        }
        else if (fromTag) {
            return calculateRepositoryUrl(tagsPath, fromTag)
        }
        else if (fromBranch) {
            return calculateRepositoryUrl(tagsPath, fromBranch)
        }
        else {
            return calculateRepositoryUrl(checkoutPath)
        }
    }

    //
    // Utility Methods
    //

    /**
     * Calculate a url within the svn repository
     * @return
     */
    protected String calculateRepositoryUrl(String... segments) {
        def fin = [svnBaseUrl];
        segments.each{
            // remove consecuative, and trailing/leading slashes from segments
            fin << it.replaceAll('/+', '/').replaceAll('^/|/$', '');
        }
        return normalizeUrl(fin.join('/'));
    }

    /**
     * Print a message to the output stream of choice.
     * @param msg
     */
    protected void println(String msg) {
        out.println(msg);
    }

    protected String getCommandText(def command) {
        final def cmdHelper = getCommandHelperBase();
        def outputText = "[error capturing output]";
        cmdHelper.runCommand(null, command){ Process proc ->
            proc.out.close() // close stdin
            // capture stdout, forward stderr
            def errThread = proc.consumeProcessErrorStream(out) // forward stderr
            try {
                outputText = proc.text.trim();
            }
            finally {
                errThread.join();
            }
        };

        // sanitize the raw output slightly and print
        def entrySepPattern = '\0|'+UNIT_SEPARATOR;
        println(outputText.replaceAll(entrySepPattern, '\n'));

        return outputText;
    }

    protected DateFormat getDateArgumentFormatter() {
        return getSvnDateFormat();
    }

    protected CommandHelper getCommandHelperBase() {
        def result = new CommandHelper(directory);
        return result;
    }

    //
    // Default Getters
    //

    /**
     * Get the scm command to use. Default to "git"
     * @return
     */
    public String getScmCommand() {
        return this.scmCommand ?: 'svn';
    }

    protected def getBaseCmd() {
        def command = [getScmCommand(), '--non-interactive']
        if (trustServerCert) {
            command << '--trust-server-cert'
        }
        if (noAuthCache) {
            command  << '--no-auth-cache'
        }
        if (svnUser) {
            command << '--username' << svnUser
            if (svnPass) {
                command << '--password' << svnPass
            }
        }
        return command;
    }

    /**
     * Normalize a url string if possible
     * @param url the normalized url
     * @return
     */
    protected String normalizeUrl(String url) {
        String result = url;
        try {
            URI uri = new URI(url);
            uri = uri.normalize();
            result = uri.toString();

            // URI mangles urls that start "file:///" - unmangle them
            String prefix = "file:///";
            if (url.startsWith(prefix)) {
                int i;
                for (i = "file:".length(); i < result.length() && result.charAt(i) == '/'; i++) {
                }
                result = prefix + result.substring(i);
            }
        }
        catch (URISyntaxException e) {
            // ignore, we'll do our best with the original value
        }
        return result;
    }

    protected def delete(File file) {
        boolean success = false;
        if (file.isFile()) {
            success = file.delete()
        }
        else if (file.isDirectory()){
            // handle directory deletion
            if (file.canonicalFile == new File('.').canonicalFile) {
                // deleting the workDir breaks child processes cause they inherit cwd
                file.eachFile{delete(it)};
            }
            else {
                success = file.deleteDir()
            }
        }
        if (!success) {
            System.err.println("Failed to remove path $file")
        }
    }
}
