/*
 * Inputs:
 *     Set<PluginSourceConfig> availableSourceConfigSet : the source configurations which should be scanned
 *     Map<String, String> triggerPropertyMap : parameters passed to the trigger invocation
 *     Map<String, String> requestPropertyMap: parameters to be placed upon the build request for matching source-configs
 *     Logger log : a logger instance
 *
 * Output:
 *     Set<PluginSourceConfig> matchingSourceConfigSet : an empty set used to return source configurations which should trigger
 *
 * Return:
 *     nothing
 */

import com.urbancode.anthill3.domain.property.PropertyValueGroupFactory;

String.metaClass.endsWithIgnoreCase = { String suffix ->
    int dl = delegate.length();
    int sl = suffix.length();
    return dl >= sl && delegate.substring(dl - sl, dl).equalsIgnoreCase(suffix);
};

// normalize URI-like strings
String.metaClass.normalize = {
    return delegate.replaceAll('(?<!:)//+', '/');
};

def parseLong = { String value ->
    return value ? value.toLong() : null
}

def resolveValue = {def module, def value ->
    return ParameterResolver.resolveForSourceConfig(value, module.getContainer());
}

def resolveModuleProperty = {def module, def propertyName ->
    return resolveValue(module, module.getPropertyValue(propertyName)?.value);
}

def calculateRepositoryUrl = { String remoteBaseUrl, String remoteUrl ->
    def result = remoteUrl;
    if (remoteBaseUrl && !(remoteUrl =~ '[a-zA-Z]+://')) {
        result = remoteBaseUrl + '/' + remoteUrl;
    }
    return result.normalize();
}

def buildSourcePropertyMap = { def module ->
    def result = [:]
    module.getPropertyMap().each{ key, value ->
        result[key] = value?.value
    }

    Long scRepoId = parseLong(result['repo']);
    if (scRepoId) {
        def pvgFactory = PropertyValueGroupFactory.getInstance();
        def repoPvg = pvgFactory.restore(scRepoId);
        result['repo'] = repoPvg.name;
        repoPvg.getPropertyMap().each{ key, value ->
            result['repo/'+key] = value?.value
        }
    }
    return result;
}

def trimSlashes = { String s ->
    def result = s.replaceFirst('^[/\\\\]+', "").replaceFirst('[/\\\\]+$', "")
    return result
}

//
//
//

String revision = triggerPropertyMap['revision']
String code = triggerPropertyMap['code']
String dirsChangedString = triggerPropertyMap['dirs-changed']
String filesChangedString = triggerPropertyMap['files-changed']
//String revisionDate = triggerPropertyMap['revision-date']
String repoPath = trimSlashes(triggerPropertyMap['repo'])

String repoName = null;
if (repoPath) {
    repoName = repoPath.split(/[\/\\]/)[-1]
}

def changedPaths = [];
if (dirsChangedString) {
    dirsChangedString.tokenize(',').each{ String dir ->
        changedPaths << trimSlashes(dir);
    };
}
if (filesChangedString) {
    filesChangedString.tokenize(',').each{ String dir ->
        changedPaths << trimSlashes(dir);
    };
}

if (changedPaths && revision) {
    for (sc in availableSourceConfigSet) {
        for (module in sc.getSourcePropertyValueGroups()) {
            def plugin = module.plugin
            if (plugin.pluginId.equals('com.urbancode.anthill3.plugin.Subversion')) {
                Map<String,String> scProps = buildSourcePropertyMap(module)
                boolean isSvnRepo = scProps['repo/isSvnRepo']
                String scRepoBasePath = trimSlashes(scProps['repo/repositoryUrl'])
                String scOffset = trimSlashes(resolveModuleProperty(module, "projectTrunkPath"))
                // full checkout url for debug logging
                String finalRepoPath = calculateRepositoryUrl(scRepoBasePath, scOffset)

                // Parse out the two sections of interest
                String scRepoName;
                String scCheckoutPath;
                if (isSvnRepo) {
                    // repo name is last part of repositoryUrl, path is projectTrunkPath
                    scRepoName = scRepoBasePath.split(/[\/\\]/)[-1]
                    scCheckoutPath = scOffset;
                }
                else {
                    // repo name is first part of projectTrunkPath, path is tail of projectTrunkPath
                    def chunks = scOffset.split(/[\/\\]/, 2);
                    scRepoName = chunks[0];
                    if (chunks.length == 2) {
                        scCheckoutPath = chunks[1];
                    }
                    else {
                        scCheckoutPath = '';
                    }
                };

                if (scRepoName != repoName) {
                    log.trace("Changed repository $repoName does not match project $sc.project.name url=$finalRepoPath,repo=$scRepoName,path=$scCheckoutPath")
                }
                else {
                    String matchingPath = changedPaths.find{ String changedPath ->
                        return changedPath.equals(scCheckoutPath) || changedPath.startsWith(scCheckoutPath+'/');
                    };
                    if (matchingPath) {
                        matchingSourceConfigSet.add(sc)
                        log.debug("Changed dir repo=$repoName,path=$matchingPath matches project $sc.project.name url=$finalRepoPath,repo=$scRepoName,path=$scCheckoutPath")
                        break;
                    }
                    else {
                        log.trace("Changed dirs $changedPaths in repo=$repoName do not match project $sc.project.name url=$finalRepoPath,repo=$scRepoName,path=$scCheckoutPath")
                    }
                }
            }
        }
    }
}
else if (!changedPaths) {
    log.warn("Could not process Svn repo trigger data: $triggerPropertyMap - no changed paths submitted (paramters dirs-changed or files-changed)")
}
else if (!revision) {
    log.warn("Could not process Svn repo trigger data: $triggerPropertyMap - no revision submitted (paramter revision")
}
else {
    log.warn("Could not process Svn repo trigger data: $triggerPropertyMap - invalid repo parameters")
}